#include "instance_tree.h"

#include "design.h"
#include "definition.h"
#include "event.h"
#include "hierarchical_channel.h"
#include "hierarchical_channel_instance.h"
#include "inoutport.h"
#include "inport.h"
#include "instance.h"
#include "object.h"
#include "outport.h"
#include "primitive_channel.h"
#include "primitive_channel_instance.h"
#include "variable.h"

#include "../tools/misc.h"
#include "../tools/variable_helper.h"
#include "../tools/function_helper.h"

/*TLM 2.0*/
#include "initiatorsocket.h"
#include "targetsocket.h"
#include "socket.h"
/*end TLM 2.0*/

risc::InstanceTree::InstanceTree()
{ }

risc::InstanceTree::InstanceTree(const InstanceTree &other)
{
  path_ = other.path_;
}

risc::InstanceTree
risc::InstanceTree::get_real_mapped_variable()
{
  InstanceTree mapped_variable;

  Element element = path_.back();
  Object *object = element.second;

  // we only consider references, so we sort all other out
  if(!object->is_reference() && !object->is_pointer()) {
    return *this;
  } else {

    InstanceTree base = *this;
    InstanceTree parent = get_parent();
    SgVariableSymbol *base_variable_symbol = NULL;
    SgVariableSymbol *parent_variable_symbol = NULL;

    base_variable_symbol = object->get_symbol();

    // basically, we are iterating as long until we get the first symbol
    // which is not a reference
    do {

      if(!parent.path_.empty()) {

        Element element = base.path_.back();

        parent_variable_symbol = get_mapped_reference(
            base_variable_symbol,
            element.first->get_ast_node(),
            parent.get_scope()->get_ast_node(),
            dynamic_cast<SgVariableDefinition*>(
              parent.get_object()->get_ast_node()));

        base = parent;
        parent = parent.get_parent();

        base_variable_symbol = parent_variable_symbol;
      }
    } while(isSgReferenceType(base_variable_symbol->get_type()));

    mapped_variable = base;

    if(dynamic_cast<Variable*>(object)) {

      mapped_variable.path_.back().second
        = base.get_scope()->find_variable(
            isSgVariableDeclaration(base_variable_symbol->get_declaration()
              ->get_declaration())->get_definition());

      // we consider this case for global variables
      if(mapped_variable.path_.back().second == NULL) {


        mapped_variable.path_.clear();
        Element e(NULL, risc::get_design()->find_global_variable(
              isSgVariableDeclaration(base_variable_symbol->get_declaration()
                ->get_declaration())->get_definition()));


        mapped_variable.path_.push_back(e);
      }
    }

    if(dynamic_cast<Event*>(object)) {

      mapped_variable.path_.back().second
        = base.path_.back().first->find_event(
            isSgVariableDeclaration(base_variable_symbol->get_declaration()
              ->get_declaration())->get_definition()) ;
    }
  }

  return mapped_variable;
}

SgVariableSymbol*
risc::InstanceTree::get_mapped_reference(
    SgVariableSymbol *reference,
    SgClassDefinition *base_class_def,
    SgClassDefinition *parent_class_def,
    SgVariableDefinition *parent_member)
{
  if(!tools::is_function_parameter_variable(reference) ) {

    SgMemberFunctionDeclaration *ctor_base
      = risc::tools::get_constructor(base_class_def);

    SgVariableSymbol *initializer_symbol = NULL;

    // First, we analyze the initializer list of the base constructor.
    // we have to figure out which variables initializes the given reference
    for(SgInitializedNamePtrList::iterator
        iter  = ctor_base->get_CtorInitializerList()->get_ctors().begin();
        iter != ctor_base->get_CtorInitializerList()->get_ctors().end();
        iter++) {

      //TODO hack
      if(reference->get_mangled_name() == (*iter)->get_mangled_name()) {

        SgAssignInitializer *assign_init
          = isSgAssignInitializer((*iter)->get_initializer());

        initializer_symbol
          = isSgVarRefExp(assign_init->get_operand())->get_symbol();

        break;
      }

      // replacement for the hack later
      // if(reference == (*iter)->search_for_symbol_from_symbol_table()) {
      //   break;
      // }
    }

    // Second, we determine the position of the initializer in the constructor
    // parameter list
    int parameter_index = 0;
    for(SgInitializedNamePtrList::iterator
        iter  = ctor_base->get_parameterList()->get_args().begin();
        iter != ctor_base->get_parameterList()->get_args().end();
        iter++, parameter_index++) {

      if(initializer_symbol == (*iter)->search_for_symbol_from_symbol_table()) {
        break;
      }
    }

    SgMemberFunctionDeclaration *ctor_parent
      = risc::tools::get_constructor(parent_class_def);

    SgConstructorInitializer *ctor_init = NULL;

    // Next, we have to determine the constructor initializer in the parent
    // class
    for(SgInitializedNamePtrList::iterator
        iter  = ctor_parent->get_CtorInitializerList()->get_ctors().begin();
        iter != ctor_parent->get_CtorInitializerList()->get_ctors().end();
        iter++) {

      //TODO hack
      if(parent_member->get_vardefn()->get_mangled_name()
          == (*iter)->get_mangled_name()) {

        ctor_init = isSgConstructorInitializer((*iter)->get_initptr());
        break;
      }
    }

    // Last, we get the mapped variable
    SgExpression *expr
      = ctor_init->get_args()->get_expressions().at(parameter_index);

    SgVarRefExp *var_ref = NULL;

    SgArrowExp *arrow_exp = isSgArrowExp(expr);
    if(arrow_exp) {
      var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());
    }

    SgVarRefExp *temp_var_ref = isSgVarRefExp(expr);
    if(temp_var_ref) {
      var_ref = temp_var_ref;
    }

    return var_ref->get_symbol();
  } else {

    SgMemberFunctionDeclaration *ctor_base
      = risc::tools::get_constructor(base_class_def);

    // First, we have to look up the
    int parameter_index = 0;
    for(SgInitializedNamePtrList::iterator
        iter  = ctor_base->get_parameterList()->get_args().begin();
        iter != ctor_base->get_parameterList()->get_args().end();
        iter++, parameter_index++) {

      if(reference == (*iter)->search_for_symbol_from_symbol_table()) {
        break;
      }
    }

    SgMemberFunctionDeclaration *ctor_parent
      = risc::tools::get_constructor(parent_class_def);

    SgConstructorInitializer *ctor_init = NULL;

    // Next, we have to determine the constructor initializer in the parent
    // class
    for(SgInitializedNamePtrList::iterator
        iter  = ctor_parent->get_CtorInitializerList()->get_ctors().begin();
        iter != ctor_parent->get_CtorInitializerList()->get_ctors().end();
        iter++) {

      //TODO hack
      if(parent_member->get_vardefn()->get_mangled_name()
          == (*iter)->get_mangled_name()) {

        ctor_init = isSgConstructorInitializer((*iter)->get_initptr());
        break;
      }
    }

    // Last, we get the mapped variable
    SgExpression *expr
      = ctor_init->get_args()->get_expressions().at(parameter_index);

    SgVarRefExp *var_ref = NULL;

    SgArrowExp *arrow_exp = isSgArrowExp(expr);
    if(arrow_exp) {
      var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());
    }

    SgVarRefExp *temp_var_ref = isSgVarRefExp(expr);
    if(temp_var_ref) {
      var_ref = temp_var_ref;
    }

    return var_ref->get_symbol();
  }
}

void
risc::InstanceTree::print_instance_tree()
{
  if(path_.empty()) {
    std::cout << "[ ]" << std::endl;

  } else {
    for(std::list<Element>::iterator
        iter  = path_.begin();
        iter != path_.end();) {

      std::cout << "[ ";

      if((*iter).first) {
        std::cout << (*iter).first->get_name();
      } else {
        std::cout << "Global Scope";
      }

      std::cout << " :: ";
      if((*iter).second) {
        std::cout << (*iter).second->get_name();

      } else {

        std::cout << "Not defined";
      }
      std::cout << " ]";

      if(++iter !=  path_.end()) {
        std::cout << " -> ";
      }
    }
    std::cout << std::endl;
  }
}

risc::InstanceTree
risc::InstanceTree::get_parent()
{
  InstanceTree parent(*this);
  if(!parent.path_.empty()) {

    parent.path_.pop_back();
    return parent;

  } else {

    return parent;
  }
}

std::list<risc::InstanceTree>
risc::InstanceTree::find_all_instances_of(SgSymbol *var_sym)
{
  Element e = path_.back();
  std::list<InstanceTree> occurence;

  Instance *inst = dynamic_cast<Instance*>(e.second);
  if(!inst) {

    if(e.second->get_ast_node()->get_vardefn()->
        search_for_symbol_from_symbol_table()
        == var_sym) {

      occurence.push_back(*this);
    }
    //std::cout << "true" << std::endl;
    return occurence;
  } else {
    //std::cout << "false" << std::endl;
    recursive_tree_walk(occurence, var_sym);

    //ZC added this return value on Oct 09 2018
    //Why this function has no return before?
    //Cant understand
    return occurence;
  }
}

void
risc::InstanceTree::recursive_tree_walk(
    std::list<risc::InstanceTree>& occurence, SgSymbol *var_sym)
{
  Instance *instance = dynamic_cast<Instance*>(path_.back().second);



  if(instance->type_ == Definition::PrimitiveChannelInstanceType) {
    return;
  }

  Class *instance_definition = instance->get_definition();

  for(ModuleInstanceVectorConstIter
      iter  = instance_definition->sub_modules_.begin();
      iter != instance_definition->sub_modules_.end();
      iter++) {

    Element sub_path;

    sub_path.first = instance->get_definition();
    sub_path.second = *iter;

    path_.push_back(sub_path);
    recursive_tree_walk(occurence, var_sym );
    path_.pop_back();
  }

  if(dynamic_cast<risc::ModuleInstance*> (instance))
    if(instance->get_ast_node()->get_vardefn()->
        search_for_symbol_from_symbol_table()
        == var_sym) 
    {
      //std::cout << instance->get_name() << " found in recursive_tree_walk " << std::endl;
      occurence.push_back(*this);
      return;
    }

  #define RECURSIVE_TREE_WALK(ITER_TYPE, VARIABLE)     \
  for( ITER_TYPE                                       \
      iter  = instance_definition-> VARIABLE .begin(); \
      iter != instance_definition-> VARIABLE .end();   \
      iter++) {                                        \
    if(risc::tools::get_sym_from_def((*iter)->get_ast_node()) == var_sym) {           \
      InstanceTree temp = *this;                       \
      Element e(instance->get_definition(), *iter);    \
      temp.path_.push_back(e);                         \
      occurence.push_back(temp);                       \
    }                                                  \
  }  

  //RECURSIVE_TREE_WALK(ModuleInstanceVectorConstIter, sub_modules_)
  RECURSIVE_TREE_WALK(PrimitiveChannelInstanceVectorConstIter, sub_prim_channel_)
  RECURSIVE_TREE_WALK(HierarchicalChannelInstanceVectorConstIter, sub_hier_channel_)
  RECURSIVE_TREE_WALK(EventVectorConstIter, events_)
  RECURSIVE_TREE_WALK(VariableVectorConstIter, variables_)
  RECURSIVE_TREE_WALK(InPortVectorConstIter, inports_)
  RECURSIVE_TREE_WALK(OutPortVectorConstIter, outports_)
  RECURSIVE_TREE_WALK(InOutPortVectorConstIter, inoutports_)

  RECURSIVE_TREE_WALK(InitiatorSocketVectorConstIter, initiatorsockets_)
  RECURSIVE_TREE_WALK(TargetSocketVectorConstIter, targetsockets_)

  #undef RECURSIVE_TREE_WALK
}

risc::InstanceTree  
risc::InstanceTree::extend_with_symbol_list(std::list<SgVariableSymbol*> syms)
{
  //std::cout << "extend tree with symbol list, size = " << syms.size() << std::endl;
  InstanceTree res;
  res = *this;
  if(syms.empty()) return res;
  risc::Module* m = dynamic_cast<Module*>(
    dynamic_cast<risc::ModuleInstance*>(get_object())->get_definition());
  //std::cout << "old m = " << m->get_name() << std::endl;
  
  for(std::list<SgVariableSymbol*>::iterator
    sym_iter  = syms.begin();
    sym_iter != syms.end();
    sym_iter ++)
  {
    //std::cout << "module name is " << m->get_name() << std::endl;
    SgVariableDefinition* mi_def = 
      isSgVariableDefinition((*sym_iter)->get_declaration()->get_definition());
    if(!mi_def) break;
    risc::Instance* mi = m->find_module(mi_def);
    if(!mi) break;
    //std::cout << "module instance name is " << mi->get_name() << std::endl;
    Element sub_path;
    sub_path.first = m;
    sub_path.second = mi;
    res.path_.push_back(sub_path);
    m = dynamic_cast<Module*>(mi->get_definition());
  }
  SgVariableDefinition* obj_def = 
      isSgVariableDefinition(syms.back()->get_declaration()->get_definition());
  if(!obj_def) {std::cout << "ERROR! obj_def==NULL" << std::endl; exit(0);}
  //std::cout << "objdef = " << SageInterface::get_name(obj_def) << std::endl;
  //std::cout << "m = " << m->get_name() << std::endl;
  risc::Object* obj = m->find_socket(obj_def);
  if(obj == NULL) {
    obj = m->find_prim_channel(obj_def);
  }
  if(obj == NULL) {
    obj = m->find_hier_channel(obj_def);
  }
  if(obj == NULL) {
    obj = m->find_port(obj_def);
  }
  if(obj == NULL) {std::cout << "ERROR! obj==NULL" << std::endl; exit(0);}
  Element sub_path;
  sub_path.first = m;
  sub_path.second = obj;
  res.path_.push_back(sub_path);

  /*std::cout << "tree extended, now is : " << std::endl;
  res.print_instance_tree();
  std::cout << std::endl;*/
  return res;
}

risc::Class*
risc::InstanceTree::get_scope()
{
  return path_.back().first;
}

risc::Object*
risc::InstanceTree::get_object()
{
  return path_.back().second;
}

SgVariableSymbol*
risc::InstanceTree::get_object_symbol()
{
  return get_object()->get_symbol();
}

risc::InstanceTree::Element
risc::InstanceTree::pop_back()
{
  Element last = path_.back();
  path_.pop_back();
  return last;
}

void
risc::InstanceTree::push_back(Element ohs)
{
  path_.push_back(ohs);
}

void
risc::InstanceTree::get_all_ports(std::list<InstanceTree>& occurence)
{
  Instance *instance = dynamic_cast<Instance*>(get_object());

  Class *definition = instance->get_definition();

  if(instance->type_ == Definition::ModuleInstanceType ||
      instance->type_ == Definition::HierarchicalChannelInstanceType) {

    for(ModuleInstanceVectorConstIter
        iter  = definition->sub_modules_.begin();
        iter != definition->sub_modules_.end();
        iter++) {

      Element sub_path;

      sub_path.first = definition;
      sub_path.second = *iter;

      path_.push_back(sub_path);
      get_all_ports(occurence);
      path_.pop_back();
    }

    for(HierarchicalChannelInstanceVectorConstIter
        iter  = definition->sub_hier_channel_.begin();
        iter != definition->sub_hier_channel_.end();
        iter++) {

      Element sub_path;

      sub_path.first = definition;
      sub_path.second = *iter;

      path_.push_back(sub_path);
      get_all_ports(occurence);
      path_.pop_back();
    }

    for(InPortVectorConstIter
        iter  = definition->inports_.begin();
        iter != definition->inports_.end();
        iter++) {

      InstanceTree temp = *this;
      Element e(definition, *iter);
      temp.path_.push_back(e);
      occurence.push_back(temp);
    }

    for(OutPortVectorConstIter
        iter  = definition->outports_.begin();
        iter != definition->outports_.end();
        iter++) {

      InstanceTree temp = *this;
      Element e(definition, *iter);
      temp.path_.push_back(e);
      occurence.push_back(temp);
    }

    for(InOutPortVectorConstIter
        iter  = definition->inoutports_.begin();
        iter != definition->inoutports_.end();
        iter++) {

      InstanceTree temp = *this;
      Element e(definition, *iter);
      temp.path_.push_back(e);
      occurence.push_back(temp);
    }
  }

  if(instance->type_ == Definition::PrimitiveChannelInstanceType) {

    for(InPortVectorConstIter
        iter  = definition->inports_.begin();
        iter != definition->inports_.end();
        iter++) {

      InstanceTree temp = *this;
      Element e(definition, *iter);
      temp.path_.push_back(e);
      occurence.push_back(temp);
    }

    for(OutPortVectorConstIter
        iter  = definition->outports_.begin();
        iter != definition->outports_.end();
        iter++) {

      InstanceTree temp = *this;
      Element e(definition, *iter);
      temp.path_.push_back(e);
      occurence.push_back(temp);
    }

    for(InOutPortVectorConstIter
        iter  = definition->inoutports_.begin();
        iter != definition->inoutports_.end();
        iter++) {

      InstanceTree temp = *this;
      Element e(definition, *iter);
      temp.path_.push_back(e);
      occurence.push_back(temp);
    }
  }
}

void
risc::InstanceTree::get_all_events(std::list<InstanceTree>& occurence)
{
  Instance *instance = dynamic_cast<Instance*>(get_object());
  Class *definition = instance->get_definition();

  if(instance->type_ == Definition::ModuleInstanceType ||
      instance->type_ == Definition::HierarchicalChannelInstanceType) {

    for(ModuleInstanceVectorConstIter
        iter  = definition->sub_modules_.begin();
        iter != definition->sub_modules_.end();
        iter++) {

      Element sub_path;

      sub_path.first = definition;
      sub_path.second = *iter;

      path_.push_back(sub_path);
      get_all_ports(occurence);
      path_.pop_back();
    }

    for(HierarchicalChannelInstanceVectorConstIter
        iter  = definition->sub_hier_channel_.begin();
        iter != definition->sub_hier_channel_.end();
        iter++) {

      Element sub_path;

      sub_path.first = definition;
      sub_path.second = *iter;

      path_.push_back(sub_path);
      get_all_ports(occurence);
      path_.pop_back();
    }

    for(EventVectorConstIter
        iter  = definition->events_.begin();
        iter != definition->events_.end();
        iter++) {

      InstanceTree temp = *this;
      Element e(definition, *iter);
      temp.path_.push_back(e);
      occurence.push_back(temp);
    }
  }

  if(instance->type_ == Definition::PrimitiveChannelInstanceType) {

    for(EventVectorConstIter
        iter  = definition->events_.begin();
        iter != definition->events_.end();
        iter++) {

      InstanceTree temp = *this;
      Element e(definition, *iter);
      temp.path_.push_back(e);
      occurence.push_back(temp);
    }
  }
}

bool
risc::InstanceTree::operator==(const InstanceTree &ohs) const
{
  if(path_.size() != ohs.path_.size()) {
    return false;
  }

  Path::const_iterator this_path_iter = path_.begin();
  Path::const_iterator ohs_path_iter = ohs.path_.begin();;

  for(; this_path_iter != path_.end(); this_path_iter++, ohs_path_iter++) {

    if(((*this_path_iter).first != (*ohs_path_iter).first) ||
        ((*this_path_iter).second != (*ohs_path_iter).second)) {
      return false;
    }
  }

  return true;
}

void
risc::InstanceTree::get_instance_tree(
    SgVariableDefinition *var_def,
    int instance_id,
    risc::InstanceTree &tree,
    VariableType type)
{
  Element e = path_.back();
  std::list<InstanceTree> occurence;

  Instance *inst = dynamic_cast<Instance*>(e.second);
  if(!inst) {

    if(e.second->get_ast_node()->get_vardefn()->search_for_symbol_from_symbol_table()
        == var_def->get_vardefn()->search_for_symbol_from_symbol_table()) {

      tree = *this;
    }
  } else {
    int current_id = 0;
    bool found = false;
    recursive_tree_walk(var_def, current_id, instance_id, tree, found, type);
  }
}

void
risc::InstanceTree::recursive_tree_walk(
    SgVariableDefinition *var_def,
    int &current_id,
    int instance_id,
    InstanceTree &tree,
    bool &found,
    VariableType type)
{
  if(found) {
    return;
  }

  Instance *instance = dynamic_cast<Instance*>(path_.back().second);

  if(instance->type_ == Definition::PrimitiveChannelInstanceType) {
    return;
  }

  Class *instance_definition = instance->get_definition();

  Element sub_path;
  path_.push_back(sub_path);
  for(ModuleInstanceVectorConstIter
      iter  = instance_definition->sub_modules_.begin();
      iter != instance_definition->sub_modules_.end();
      iter++) {


    path_.back().first = instance->get_definition();
    path_.back().second = *iter;

    recursive_tree_walk(var_def, current_id, instance_id, tree, found, type);
  }
  path_.pop_back();

  #define RECURSIVE_TREE_WALK(ITER_TYPE, VARIABLE, VARIABLE_TYPE) \
  if(type == VARIABLE_TYPE) {                                     \
    for( ITER_TYPE                                                \
        iter  = instance_definition-> VARIABLE .begin();          \
        iter != instance_definition-> VARIABLE .end();            \
        iter++) {                                                 \
      if((*iter)->get_ast_node() == var_def) {                    \
        if(current_id == instance_id) {                           \
          InstanceTree temp = *this;                              \
          Element e(instance->get_definition(), *iter);           \
          temp.path_.push_back(e);                                \
          tree = temp;                                            \
          found = true;                                           \
          return;                                                 \
        }                                                         \
        current_id++;                                             \
      }                                                           \
    }                                                             \
  }

  RECURSIVE_TREE_WALK(ModuleInstanceVectorConstIter, sub_modules_, ModuleInstance)
  RECURSIVE_TREE_WALK(PrimitiveChannelInstanceVectorConstIter, sub_prim_channel_, PrimitiveChannelInstance)
  RECURSIVE_TREE_WALK(HierarchicalChannelInstanceVectorConstIter, sub_hier_channel_, HierarchicalChannelInstance)
  RECURSIVE_TREE_WALK(EventVectorConstIter, events_, EventInstance)
  RECURSIVE_TREE_WALK(VariableVectorConstIter, variables_, VariableInstance)
  RECURSIVE_TREE_WALK(InPortVectorConstIter, inports_, PortInstance)
  RECURSIVE_TREE_WALK(OutPortVectorConstIter, outports_, PortInstance)
  RECURSIVE_TREE_WALK(InOutPortVectorConstIter, inoutports_, PortInstance)

  #undef RECURSIVE_TREE_WALK
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
