#include "internal_representation_helper.h"

#include "rose.h"

#include <string>

#include "design.h"
#include "hierarchical_channel.h"
#include "module.h"

#include "../tools/misc.h"

risc::Module*
risc::find_module_definition(std::string name)
{
  for(ModuleVectorConstIter
      module_iter  = risc::get_design()->module_definitions_.begin();
      module_iter != risc::get_design()->module_definitions_.end();
      module_iter++) {

    std::string demangled_name =
      (*module_iter)->get_ast_node()->get_declaration()->get_type()
        ->get_mangled();

    risc::tools::find_and_replace_all(demangled_name, "\"", "");
    risc::tools::find_and_replace_all(demangled_name, "__scope__", "::");
    risc::tools::find_and_replace_all(demangled_name, "class_", "");

    if(demangled_name == name) {
      return *module_iter;
    }
  }

  // If we are here something went wrong
  return NULL;
}

risc::HierarchicalChannel*
risc::find_hierarchical_channel_definition(std::string name)
{
  for(HierarchicalChannelVectorConstIter
      channel_iter  = risc::get_design()->hierarchical_channel_.begin();
      channel_iter != risc::get_design()->hierarchical_channel_.end();
      channel_iter++) {

    std::string demangled_name
      = SageInterface::get_name(
          (*channel_iter)->get_ast_node()->get_declaration());

    risc::tools::find_and_replace_all(demangled_name, "\"", "");
    risc::tools::find_and_replace_all(demangled_name, "__scope__", "::");
    risc::tools::find_and_replace_all(demangled_name, "class_", "");

    if(demangled_name == name) {
      return *channel_iter;
    }
  }

  // If we are here something went wrong
  return NULL;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
