#ifndef PORT_CALL_PATH_H_INCLUDED_
#define PORT_CALL_PATH_H_INCLUDED_

#include "rose.h"

namespace risc {

namespace sg {

class PortCallPath {

public:
  /**
   * \brief   First type representation of PCP.
   *          A list of of function calls;
   */
  std::list<SgFunctionCallExp*> path_;

  //2018 07 18 ZC
  //for pcp+psg, details refer to change_logs_ZC 1)
  /**
   * \brief   Second type representation of PCP.
   *          A list of of sockets/ports;
   */
  std::list<SgVariableDefinition*> port_;

  /**
   * \brief   Type of PCP
   */
  int pcp_type_; 

  PortCallPath();
  PortCallPath(std::list<SgVariableDefinition*>);
  ~PortCallPath();

  /**
   * \fn    print
   * \brief Prints the PCP
   */
  void print();
};

bool operator==(const PortCallPath &lhs , const PortCallPath &rhs);
bool operator<(const PortCallPath &lhs , const PortCallPath &rhs); // (RD, 11/22/19)

struct SymbolWithPath {
  SgVariableSymbol *symbol_;
  PortCallPath pcp_;

  SymbolWithPath(SgVariableSymbol* symbol, PortCallPath pcp):
    symbol_(symbol),
    pcp_(pcp)
  { }

  bool operator<(const SymbolWithPath &other) const // needed for std::set (RD, 11/22/18)
  {
  if(symbol_ != other.symbol_) {
    return (symbol_ < other.symbol_);  /* pointer comparison */
    }
  return (pcp_ < other.pcp_);
  }
};

struct NodeWithPath {
  SgNode *node_;
  PortCallPath pcp_;

  NodeWithPath(SgNode* node, PortCallPath pcp):
    node_(node),
    pcp_(pcp)
  { }
};

} // end of namespace sg

} // end of namespace risc

#endif /* PORT_CALL_PATH_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
