#ifndef PREDICTION_TIME_ADVANCE_TABLE_H_INCLUDED_
#define PREDICTION_TIME_ADVANCE_TABLE_H_INCLUDED_

#include<string>

#include "time.h"
#include "segment_graph.h"

namespace risc {

namespace sg {

class PredictionTimeAdvanceTable {

public:

  PredictionTimeAdvanceTable(risc::sg::SegmentGraph &sg, int prediction_steps);

  /**
   * \fn    print
   * \brief This function prints the prediction time advance table in the
   *        terminal
   */
  void print();

  /**
   * \fn    print(std::string filename)
   * \brief Generates a html file of the prediction time advance table
   */
  void print(std::string filename);

  /**
   * \fn index_to_segment_id(int index)
   * \brief translates a given index to the related segment id
   */
  int index_to_segment_id(int index);

  /**
   * \fn    segment_id_to_index(int segment_id)
   * \brief translates a given segment id to the related index
   */
  int segment_id_to_index(int segment_id);

  /**
   * \brief    The actual table
   */
  std::vector<std::vector< risc::sg::Time> > table_;

  /**
   * \brief    A mapping from index to segment id
   */
  std::vector<int> index_to_segment_;

  /**
   * \brief    The max steps to be predicted
   */
  int prediction_steps_;

private:

  /**
   * \fn    compute_table
   * \brief This function computes the prediction time advance table
   */
  void compute_table();

  /**
   * \fn    analyze(VertexDescriptor start, int steps)
   * \brief This function computes the prediction time advance table
   */
  void analyze(VertexDescriptor start, int steps);

  /**
   * \brief    Reference to the actual segment graph
   */
  SegmentGraph &sg_;
};

} // end of namespace sg

} // end of namespace risc

#endif /* PREDICTION_TIME_ADVANCE_TABLE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
