#ifndef SEARCH_CLASS_DEFINITION_BY_NAME_H_INCLUDED_
#define SEARCH_CLASS_DEFINITION_BY_NAME_H_INCLUDED_

#include <string>
#include "rose.h"

namespace risc
{

namespace tools
{

/**
 * \brief                      This function searchs for a class by name
 *                             (string) in namespace (string)
 * \param  class_name          Name of the class
 * \param  qualified_namespace The absolute qualified namespace
 * \return                     Pointer to the class definition. The pointer is
 *                             zero, if the class does not exist in the given
 *                             namespace
 */
SgClassDefinition*
find_class_definition(std::string class_name, std::string qualified_namespace);

/**
 * \brief                      This function searchs for a typedef's by name
 *                             (string) in namespace (string)
 * \param  typedef_name          Name of the class
 * \param  qualified_namespace The absolute qualified namespace
 * \return                     Pointer to the class definition. The pointer is
 *                             zero, if the class does not exist in the given
 *                             namespace
 */
SgTypedefDeclaration* find_typedef_declaration(
    std::string typedef_name,
    std::string qualified_namespace);

} // end namespace tools

} // end namespace risc

#endif /* SEARCH_CLASS_DEFINITION_BY_NAME_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
