#include <stdio.h>
#include "systemc.h"
#include "tlm.h"                                    // TLM headers
#include "tlm_utils/simple_initiator_socket.h"      // TLM simple initiator socket
#include "tlm_utils/simple_target_socket.h"
#include <string>
using namespace sc_core;
using namespace std;


// frame rates
#define AUDIO_DELAY     26120, SC_US
#define VIDEO_DELAY     33330, SC_US
#define HALF_AUDIO_DELAY 26120/2, SC_US
#define HALF_VIDEO_DELAY 33330/2, SC_US

// simulated time
#define CLIP_LENGTH     10, SC_SEC

// simulated workload unit
#define AUDIOLOAD        7500000
#define VIDEOLOAD       10000000

#define BUFSIZE		1


SC_MODULE(TLM2_Fifo)
{
  public:
    tlm_utils::simple_target_socket<TLM2_Fifo>  in; //in port of the fifo
    tlm_utils::simple_target_socket<TLM2_Fifo>  out; //out port of the fifo
    /*SC_CTOR(TLM2_Fifo)
    {
      n=0;
      out.register_b_transport(this, &TLM2_Fifo::read);
      in.register_b_transport(this, &TLM2_Fifo::write);
    }

    void read(tlm::tlm_generic_payload& trans, sc_time& delay)
    {
      unsigned char*   ptr_read = trans.get_data_ptr();
      unsigned int     len = trans.get_data_length();
      while(n==0){
        wait(r);
      }
      n=0;
      memcpy(ptr_read, &buffer, len);
      s.notify(SC_ZERO_TIME);
    }

    void write(tlm::tlm_generic_payload& trans, sc_time& delay)
    {
      unsigned char*   ptr_write = trans.get_data_ptr();
      unsigned int     len = trans.get_data_length();
      while(n==1){
        wait(s);
      }
      n=1;
      memcpy(&buffer, ptr_write, len);
      r.notify(SC_ZERO_TIME);
    }*/
    
    SC_CTOR(TLM2_Fifo)
    {
      wr = 0;
      ws = 0;
      p = 0;
      n = 0;
      m_size = BUFSIZE; 
      buffer = new long long[BUFSIZE]; 
      out.register_b_transport(this, &TLM2_Fifo::read);
      in.register_b_transport(this, &TLM2_Fifo::write);
      
    }

    void read(tlm::tlm_generic_payload& trans, sc_time& delay)
    {
      unsigned char*   ptr_read = trans.get_data_ptr();
      unsigned int     len = trans.get_data_length();
      
      //std::cout << len << std::endl;

      long long d;
      while (!n)
      {
        wr ++;
        wait(r);
        wr --;
      }

      if (n <= p)
      {
        d = buffer[p - n];
      }
      else
      {
        d = buffer[p + m_size - n];
      }
      n --;

      if (ws)
      {
        s.notify(SC_ZERO_TIME);
      }

      memcpy(ptr_read, &d, len);

    }

    void write(tlm::tlm_generic_payload& trans, sc_time& delay)
    {
      unsigned char*   ptr_write = trans.get_data_ptr();
      unsigned int     len = trans.get_data_length();
      //std::cout << len << std::endl;

      long long d;
      memcpy(&d, ptr_write, len);

      while (n >= m_size)
      {
        ws ++;
        wait(s);
        ws --;
      }

      buffer[p] = d;
      p ++;
      if (p >= m_size)
      {
        p = 0;
      }
      n ++;

      if (wr)
      {
        r.notify(SC_ZERO_TIME);
      }
    }

  private:
    sc_event       r, s;
    unsigned long  wr;
    unsigned long  ws;
    unsigned long  p;
    unsigned long  n;
    long long      *buffer;
    //long long      buffer;
    unsigned long  m_size;
};

SC_MODULE(Stimulus)
{
  tlm_utils::simple_initiator_socket<Stimulus> AudioStreamL;
  tlm_utils::simple_initiator_socket<Stimulus> AudioStreamR;
  long long audioLeft;
  long long audioRight;

  void AudioL()
  {

    sc_time ClipLength(CLIP_LENGTH);
    sc_time AudioDelay(AUDIO_DELAY);
    sc_time t2(0, SC_US);
    while(t2 < ClipLength)
    {
      t2 += AudioDelay;
      sc_time delay1 = sc_time(0, SC_NS);
      tlm::tlm_generic_payload transl;
      transl.set_data_ptr( reinterpret_cast<unsigned char*>(&audioLeft) );
      transl.set_data_length( 8 );
      AudioStreamL->b_transport(transl, delay1);
    }
  }

  void AudioR()
  {

    sc_time ClipLength(CLIP_LENGTH);
    sc_time AudioDelay(AUDIO_DELAY);
    sc_time t2(0, SC_US);
    while(t2 < ClipLength)
    {
      t2 += AudioDelay;
      sc_time delay2 = sc_time(0, SC_NS);
      tlm::tlm_generic_payload transr; 
      transr.set_data_ptr( reinterpret_cast<unsigned char*>(&audioRight) );
      transr.set_data_length( 8 );
      AudioStreamR->b_transport(transr, delay2);
    }
  }
  
  SC_CTOR(Stimulus)
  {
    audioLeft = 27;
    audioRight = 15;
    SC_THREAD(AudioL);
    SC_THREAD(AudioR);
  }
};

SC_MODULE(AudioCodec)
{
  tlm_utils::simple_initiator_socket<AudioCodec> p1;
  long long d, inFrame, outFrame;

  long long decode(long long x)
  {
    int i;

    for (i = 0; i < AUDIOLOAD; i++)
    {
      d *= x;
    }
    return d;
  }
  
  void main(void)
  {

    d = 1;

    while(1)
    {
      //read
      tlm::tlm_generic_payload transac; 
      sc_time delay0 = sc_time(0, SC_NS);
      transac.set_data_ptr( reinterpret_cast<unsigned char*>(&inFrame) );
      transac.set_data_length( 8 );
      p1->b_transport(transac, delay0);

      wait(HALF_AUDIO_DELAY);
      outFrame = decode(inFrame);
      std::cout << inFrame << " " ;
      fflush(stdout);
      wait(HALF_AUDIO_DELAY);
     
    }
  }

  SC_CTOR(AudioCodec)
  {
    SC_THREAD(main);
  }
};


SC_MODULE(Top)
{
  Stimulus       stimulus;
  AudioCodec     left;
  AudioCodec     right;

  TLM2_Fifo      fifo_S_L, fifo_S_R;
  void before_end_of_elaboration()
  {
    
    stimulus.AudioStreamL.bind(fifo_S_L.in);
    stimulus.AudioStreamR.bind(fifo_S_R.in);

    left.p1.bind(fifo_S_L.out);
    right.p1.bind(fifo_S_R.out);
  }

  SC_CTOR(Top)
  : stimulus("stimulus")
  , left("left")
  , right("right")
  , fifo_S_L("fifo_S_L")
  , fifo_S_R("fifo_S_R")
  { }
};

int sc_main(int argc, char* argv[])
{
  Top top("top");

  printf("Playing...\n");
  sc_start();
  printf("Done!\n");
  return 0;
}

// EOF
