// play.cpp: conceptual audio/video stream decoding example
// Author: RISC project members, CECS
// 
// 27/11/18 ZC  use sockets to replace ports and channels
// 06/05/18 RD	confirmed clean sc_time constants with GCC 6.3.1
// 04/27/18 RD  clean sc_time constants for newer compilers (e.g. GCC 4.9.2)
// 04/26/18 RD  merged tracing version into main file (define TRACE_FILE)
// 04/24/18 DM  added tracing of stream data (sc_trace is now supported)
// 07/10/17 RD  split the delay into two parts (as suggested by ZC)
// 04/05/17 RD  removed sc_stop from stimulus, let monitor terminate;
//              added COUNT_FRAMES to ensure proper completion
// 03/08/17 RD  avoid duplicated DELAY macros, set BUFSIZE to 10
// 01/13/17 GL  separate version play_fifo.cpp with sc_fifo channels
// 10/04/16 RD  removed obsolete C artifacts from the user-defined queue channel
// 09/30/16 RD  cleanup for release 0.3.0
// 03/07/15 RD  bug fix: use sc_stop() to quit cleanly (exit() is not MT-safe)
// 10/30/15 RD  polished for release 0.2.1
// 09/04/15 GL  use 'long long' (not 'double') to avoid floating-point overflow
//              (FPU exception handling shows high performance variation;
//              moreover, integer arithmetic is more realistic for decoder)
// 08/30/15 GL  initial version based on SpecC example play.sc by RD


#include <stdio.h>
#include "systemc.h"
#include "tlm.h"                                    // TLM headers
#include "tlm_utils/simple_initiator_socket.h"      // TLM simple initiator socket
#include "tlm_utils/simple_target_socket.h"
#include <string>
using namespace sc_core;
using namespace std;

// frame rates
#define AUDIO_DELAY     26120, SC_US
#define VIDEO_DELAY     33330, SC_US
#define HALF_AUDIO_DELAY 26120/2, SC_US
#define HALF_VIDEO_DELAY 33330/2, SC_US

// simulated time
#define CLIP_LENGTH     1, SC_SEC

// simulated workload unit
#define AUDIOLOAD        7500000
#define VIDEOLOAD       10000000

// illustrate and check frame decoding
#define PRINT_TIME
#define PRINT_FRAMES
#define COUNT_FRAMES
//#define TRACE_FILE

// buffer size in queue channels
#define BUFSIZE		1


SC_MODULE(TLM2_Fifo)
{
  public:
    tlm_utils::simple_target_socket<TLM2_Fifo>  in; //in port of the fifo
    tlm_utils::simple_target_socket<TLM2_Fifo>  out; //out port of the fifo
    SC_CTOR(TLM2_Fifo)
    {
      n=0;
      out.register_b_transport(this, &TLM2_Fifo::read);
      in.register_b_transport(this, &TLM2_Fifo::write);
    }

    void read(tlm::tlm_generic_payload& trans, sc_time& delay)
    {
      unsigned char*   ptr_read = trans.get_data_ptr();
      unsigned int     len = trans.get_data_length();
      string _name = name();
      bool should_print = false;//(_name == "top.channel12_audioCodecR_to_SpeakerR" || _name == "top.channel1_stimulusR_to_audioCodecR");
      
      if(should_print) cout << "\nentering " << _name << " :: read, n= " << n << endl<< endl;
      if(n==0){
        if(should_print) cout << "\nin " << _name << " :: read, since n==0, waiting for event r"<< endl<< endl;
        wait(r);
        if(should_print) cout << "\nin " << _name << " :: read, event r notified, wake up"<< endl<< endl;
      }
      memcpy(ptr_read, &buffer, len);
      n=0;
      if(should_print) cout << "\nin " << _name << " :: read, n set to " << n << endl<< endl;
      s.notify(SC_ZERO_TIME);
      if(should_print) cout << "\nin " << _name << " :: read, s notified " << endl<< endl;
    }

    void write(tlm::tlm_generic_payload& trans, sc_time& delay)
    {
      unsigned char*   ptr_write = trans.get_data_ptr();
      unsigned int     len = trans.get_data_length();
      string _name = name();bool should_print = false;//(_name == "top.channel12_audioCodecR_to_SpeakerR" || _name == "top.channel1_stimulusR_to_audioCodecR");
      
      if(should_print) cout << "\nentering " << _name << " :: write, n= " << n << endl<< endl;
      if(n==1){
        if(should_print) cout << "\nin " << _name << " :: write, since n==1, waiting for event s"<< endl<< endl;
        wait(s);
        if(should_print) cout << "\nin " << _name << " :: write, event s notified, wake up"<< endl<< endl;
      }
      
      memcpy(&buffer, ptr_write, len);
      n=1;
      
      if(should_print) cout << "\nin " << _name << " :: write, n set to " << n << endl<< endl;
      r.notify(SC_ZERO_TIME);
      if(should_print) cout << "\nin " << _name << " :: write, r notified " << endl<< endl;
    }
    
    /*
    SC_CTOR(TLM2_Fifo)
    {
      wr = 0;
      ws = 0;
      p = 0;
      n = 0;
      m_size = BUFSIZE; 
      buffer = new long long[BUFSIZE]; 
      out.register_b_transport(this, &TLM2_Fifo::read);
      in.register_b_transport(this, &TLM2_Fifo::write);
      
    }

    void read(tlm::tlm_generic_payload& trans, sc_time& delay)
    {

      unsigned char*   ptr_read = trans.get_data_ptr();
      unsigned int     len = trans.get_data_length();
      
      //std::cout << len << std::endl;

      long long d;
      cout << name() << ":: read, n= " << n << endl;
      while (!n)
      {
        wr ++;
        wait(r);
        wr --;
      }

      if (n <= p)
      {
        d = buffer[p - n];
      }
      else
      {
        d = buffer[p + m_size - n];
      }
      n --;

      if (ws)
      {
        s.notify(SC_ZERO_TIME);
      }

      memcpy(ptr_read, &d, len);

    }

    void write(tlm::tlm_generic_payload& trans, sc_time& delay)
    {
      unsigned char*   ptr_write = trans.get_data_ptr();
      unsigned int     len = trans.get_data_length();
      //std::cout << len << std::endl;

      long long d;
      memcpy(&d, ptr_write, len);
      cout << name() << ":: write, n= " << n << endl;

      while (n >= m_size)
      {
        ws ++;
        wait(s);
        ws --;
      }

      buffer[p] = d;
      p ++;
      if (p >= m_size)
      {
        p = 0;
      }
      n ++;

      if (wr)
      {
        r.notify(SC_ZERO_TIME);
      }
    }*/

  private:
    sc_event       r, s;
    unsigned long  wr;
    unsigned long  ws;
    unsigned long  p;
    unsigned long  n;
    //long long      *buffer;
    long long      buffer;
    unsigned long  m_size;
};

SC_MODULE(Stimulus)
{
  // sc_fifo_out<long long> VideoStream;
  // sc_fifo_out<long long> AudioStreamL;
  // sc_fifo_out<long long> AudioStreamR;
  tlm_utils::simple_initiator_socket<Stimulus> VideoStream;
  tlm_utils::simple_initiator_socket<Stimulus> AudioStreamL;
  tlm_utils::simple_initiator_socket<Stimulus> AudioStreamR;
  long long video;
  long long audioLeft;
  long long audioRight;


  void Video()
  {

    sc_time ClipLength(CLIP_LENGTH);
    sc_time VideoDelay(VIDEO_DELAY);
    sc_time t1(0, SC_US);
    while(t1 < ClipLength){
      t1 += VideoDelay;
      sc_time delay0 = sc_time(0, SC_NS);
      tlm::tlm_generic_payload trans0; 
      trans0.set_data_ptr( reinterpret_cast<unsigned char*>(&video) );
      trans0.set_data_length( 8 );
      VideoStream->b_transport(trans0, delay0);
    }
  }

  void AudioL()
  {

    sc_time ClipLength(CLIP_LENGTH);
    sc_time AudioDelay(AUDIO_DELAY);
    sc_time t2(0, SC_US);
    while(t2 < ClipLength)
    {
      t2 += AudioDelay;
      sc_time delay1 = sc_time(0, SC_NS);
      tlm::tlm_generic_payload trans1;
      trans1.set_data_ptr( reinterpret_cast<unsigned char*>(&audioLeft) );
      trans1.set_data_length( 8 );
      AudioStreamL->b_transport(trans1, delay1);
    }
  }

  void AudioR()
  {

    sc_time ClipLength(CLIP_LENGTH);
    sc_time AudioDelay(AUDIO_DELAY);
    sc_time t2(0, SC_US);
    while(t2 < ClipLength)
    {
      t2 += AudioDelay;
      sc_time delay2 = sc_time(0, SC_NS);
      tlm::tlm_generic_payload trans2; 
      trans2.set_data_ptr( reinterpret_cast<unsigned char*>(&audioRight) );
      trans2.set_data_length( 8 );
      //cout << "\nStimulus AudioR enters channel channel1_stimulusR_to_audioCodecR\n" << endl;
      AudioStreamR->b_transport(trans2, delay2);
      //cout << "\nStimulus AudioR exits channel channel1_stimulusR_to_audioCodecR\n" << endl;
    }
  }
  
  SC_CTOR(Stimulus)
  {
    video = 3;
    audioLeft = 27;
    audioRight = 15;
    SC_THREAD(Video);
    SC_THREAD(AudioL);
    SC_THREAD(AudioR);
  }
};

SC_MODULE(VideoCodec)
{
  tlm_utils::simple_initiator_socket<VideoCodec> p1;
  tlm_utils::simple_initiator_socket<VideoCodec> p2;

  long long d, outFrame, inFrame, counter;
  
  long long decode(long long x)
  {
    int i;

    for (i = 0; i < VIDEOLOAD; i++)
    {
      d *= x;
    }
    return d;
  }
  
  
  void main(void)
  {

    d = 1;
    counter = 0;
    while(1)
    {
      //read
      tlm::tlm_generic_payload trans0; 
      sc_time delay0 = sc_time(0, SC_NS);
      trans0.set_data_ptr( reinterpret_cast<unsigned char*>(&inFrame) );
      trans0.set_data_length( 8 );
      p1->b_transport(trans0, delay0);

      wait(HALF_VIDEO_DELAY);
      //wait(VIDEO_DELAY);
      outFrame = decode(inFrame);
      ++counter;
      wait(HALF_VIDEO_DELAY);
      

      //write
      tlm::tlm_generic_payload trans1; 
      sc_time delay1 = sc_time(0, SC_NS);
      trans1.set_data_ptr( reinterpret_cast<unsigned char*>(&outFrame) );
      trans1.set_data_length( 8 );
      p2->b_transport(trans1, delay1);
    }
  }

  SC_CTOR(VideoCodec)
  {
    SC_THREAD(main);
  }
};

SC_MODULE(AudioCodec)
{
  tlm_utils::simple_initiator_socket<AudioCodec> p1;
  tlm_utils::simple_initiator_socket<AudioCodec> p2;
  long long d, inFrame, outFrame, counter;
  int id;
  long long decode(long long x)
  {
    int i;

    for (i = 0; i < AUDIOLOAD; i++)
    {
      d *= x;
    }
    return d;
  }
  
  void main(void)
  {

    d = 1;
    counter = 0;

    while(1)
    {
      //read
      tlm::tlm_generic_payload trans0; 
      sc_time delay0 = sc_time(0, SC_NS);
      trans0.set_data_ptr( reinterpret_cast<unsigned char*>(&inFrame) );
      trans0.set_data_length( 8 );
      //if(id==1) cout << "\nAudioCodec Right enters channel channel1_stimulusR_to_audioCodecR\n" << endl;
      p1->b_transport(trans0, delay0);
      //if(id==1) cout << "\nAudioCodec Right exits channel channel1_stimulusR_to_audioCodecR\n" << endl;

      wait(HALF_AUDIO_DELAY);
      //wait(AUDIO_DELAY);
      outFrame = decode(inFrame);
      ++counter;
      wait(HALF_AUDIO_DELAY);
     
      //write
      tlm::tlm_generic_payload trans1; 
      sc_time delay1 = sc_time(0, SC_NS);
      trans1.set_data_ptr( reinterpret_cast<unsigned char*>(&outFrame) );
      trans1.set_data_length( 8 );
      //if(id==1) cout << "\nAudioCodec Right enters channel channel12_audioCodecR_to_SpeakerR\n" << endl;
      p2->b_transport(trans1, delay1);
      //if(id==1) cout << "\nAudioCodec Right exits channel channel12_audioCodecR_to_SpeakerR\n" << endl;

    }
  }

  AudioCodec(sc_module_name name, int _id): sc_module(name), id(_id)
  {
    SC_THREAD(main);
  }
  SC_HAS_PROCESS(AudioCodec);
};

SC_MODULE(DUT)
{
  VideoCodec video;
  AudioCodec left, right;
  SC_CTOR(DUT): video("video"), left("left",0), right("right",1)
  { }
};

SC_MODULE(Display)
{
  tlm_utils::simple_initiator_socket<Display> p1;
  long long Frame;
#ifdef COUNT_FRAMES
  unsigned FramesDisplayed;
#endif

  void display(long long /* Frame */)
  {
#ifdef PRINT_FRAMES
    cout << "V" ;       // display the frame
    fflush(stdout);
#endif
#ifdef COUNT_FRAMES
    FramesDisplayed++;
#endif
  }
  
  void main(void)
  {
    
    while(1)
    {
      tlm::tlm_generic_payload trans0; 
      sc_time delay0 = sc_time(0, SC_NS);
      trans0.set_data_ptr( reinterpret_cast<unsigned char*>(&Frame) );
      trans0.set_data_length( 8 );
      p1->b_transport(trans0, delay0);

      display(Frame);
    }
  }

  SC_CTOR(Display)
  {
#ifdef COUNT_FRAMES
    FramesDisplayed = 0;
#endif
    SC_THREAD(main);
  }
  
};

SC_MODULE(Speaker)
{
  tlm_utils::simple_initiator_socket<Speaker> p1;

  long long Frame;
#ifdef COUNT_FRAMES
  unsigned FramesPlayed;
#endif

  char Channel;

  void sound(long long /* Frame */)
  {
#ifdef PRINT_FRAMES
    cout << Channel; // output the sound
    fflush(stdout);
#endif
#ifdef COUNT_FRAMES
    FramesPlayed++;
#endif
  }

  
  void main(void)
  {
    long long Frame;

    while(1)
    {
      tlm::tlm_generic_payload trans0; 
      sc_time delay0 = sc_time(0, SC_NS);
      trans0.set_data_ptr( reinterpret_cast<unsigned char*>(&Frame) );
      trans0.set_data_length( 8 );
      //if(Channel == 'R') cout << "\nSpeaker Right enters channel channel12_audioCodecR_to_SpeakerR\n" << endl;
      p1->b_transport(trans0, delay0);
      //if(Channel == 'R') cout << "\nSpeaker Right exits channel channel12_audioCodecR_to_SpeakerR\n" << endl;

      sound(Frame);
    }
  }

  Speaker(sc_module_name name, char chn): sc_module(name), Channel(chn)
  {
    
#ifdef COUNT_FRAMES
    FramesPlayed = 0;
#endif
    SC_THREAD(main);
  }
  SC_HAS_PROCESS(Speaker);
};

SC_MODULE(Monitor)
{
  Display screen;
  Speaker speakerL, speakerR;

  SC_CTOR(Monitor):
        screen("screen"),
        speakerL("speakerL", 'L'),
        speakerR("speakerR", 'R')
  { }
};

SC_MODULE(Top)
{
  Stimulus       stimulus;
  DUT            dut;
  Monitor        monitor;
  TLM2_Fifo      fifo_S_V, fifo_S_AR, fifo_S_AL;
  TLM2_Fifo      fifo_V_S, fifo_AR_SR, fifo_AL_SL;
  void before_end_of_elaboration()
  {
    stimulus.VideoStream.bind(fifo_S_V.in);
    stimulus.AudioStreamL.bind(fifo_S_AL.in);
    stimulus.AudioStreamR.bind(fifo_S_AR.in);

    dut.video.p1.bind(fifo_S_V.out);
    dut.left.p1.bind(fifo_S_AL.out);
    dut.right.p1.bind(fifo_S_AR.out);

    dut.video.p2.bind(fifo_V_S.in);
    dut.left.p2.bind(fifo_AL_SL.in);
    dut.right.p2.bind(fifo_AR_SR.in);

    monitor.screen.p1.bind(fifo_V_S.out);
    monitor.speakerL.p1.bind(fifo_AL_SL.out);
    monitor.speakerR.p1.bind(fifo_AR_SR.out);

    
  }

  SC_CTOR(Top)
  : stimulus("stimulus")
  , dut("dut")
  , monitor("monitor")
  , fifo_S_V("fifo_S_V")
  , fifo_S_AR("channel1_stimulusR_to_audioCodecR")
  , fifo_S_AL("fifo_S_AL")
  , fifo_V_S("fifo_V_S")
  , fifo_AR_SR("channel12_audioCodecR_to_SpeakerR")
  , fifo_AL_SL("fifo_AL_SL")
  { }
};

int sc_main(int argc, char* argv[])
{
  Top top("top");
  sc_time	ClipLength(CLIP_LENGTH);
  sc_time	VideoDelay(VIDEO_DELAY);
  sc_time	AudioDelay(AUDIO_DELAY);
#ifdef TRACE_FILE
  sc_trace_file *f = sc_create_vcd_trace_file(argv[0]);
  f->set_time_unit(sc_time(10, SC_US).to_seconds(), SC_SEC);
  sc_trace(f, top.dut.video.d, "top.dut.video.d");
  sc_trace(f, top.dut.left.d, "top.dut.left.d");
  sc_trace(f, top.dut.right.d, "top.dut.right.d");
#  ifdef COUNT_FRAMES
  sc_trace(f, top.monitor.screen.FramesDisplayed, "top.monitor.screen.FramesDisplayed");
  sc_trace(f, top.monitor.speakerL.FramesPlayed, "top.monitor.speakerL.FramesPlayed");
  sc_trace(f, top.monitor.speakerR.FramesPlayed, "top.monitor.speakerR.FramesPlayed");
#  endif
#endif

#ifdef PRINT_TIME
  printf("%s: ", sc_time_stamp().to_string().c_str());
#endif
  printf("Playing...\n");
  sc_start();
#ifdef PRINT_FRAMES
  printf("\n");
#endif
#ifdef COUNT_FRAMES
  printf("%u video frames displayed\n", top.monitor.screen.FramesDisplayed);
  printf("%u left audio frames played\n", top.monitor.speakerL.FramesPlayed);
  printf("%u right audio frames played\n", top.monitor.speakerR.FramesPlayed);
  printf("\n%u video frames calculated\n", top.dut.video.counter);
  printf("%u left audio frames calculated\n", top.dut.left.counter);
  printf("%u right audio frames calculated\n", top.dut.right.counter);
  assert(top.monitor.screen.FramesDisplayed ==
		ClipLength.value()/VideoDelay.value()+1);
  assert(top.monitor.speakerL.FramesPlayed ==
		ClipLength.value()/AudioDelay.value()+1);
  assert(top.monitor.speakerR.FramesPlayed ==
		ClipLength.value()/AudioDelay.value()+1); 
#endif
#ifdef PRINT_TIME
  printf("%s: ", sc_time_stamp().to_string().c_str());
  assert(sc_time_stamp() >= ClipLength);
#endif
#ifdef TRACE_FILE
  sc_close_vcd_trace_file(f);
#endif
  printf("Done!\n");
  return 0;
}

// EOF
