#include "systemc.h"

class Interface: virtual public sc_interface
{
public:
  virtual int read(void) = 0;
  virtual void write(const int) = 0;
};

class Channel:
  public sc_channel,
  public Interface
{
public:
  Channel(sc_module_name name):
    sc_channel(name)
  {
      v = false;
      w = false;
  }

  int read(void)
  {
    if (!v) {
      w = true;
      wait(req);
      w = false;
    }
    int temp = tmpd;
    v = false;
    ack.notify(SC_ZERO_TIME);
    wait(ack);
    return temp;
  }

  void write(const int X)
  {
    tmpd = X;
    v = true;
    if (w)
    {
      req.notify(SC_ZERO_TIME);
    }
    wait(ack);
  }

private:
  sc_event req;
  sc_event ack;
  bool v, w;
  int tmpd;
};

class SubModule1: public sc_module {

public:

  SC_HAS_PROCESS(SubModule1);

  SubModule1(sc_module_name mn):
    sc_module(mn)
  { }
};


class SubModule2: public sc_module {

public:

  SC_HAS_PROCESS(SubModule2);
  SubModule1 *submodule1;

  SubModule2(sc_module_name mn):
    sc_module(mn)
  {
    submodule1 = new SubModule1("submodule1");
  }
};


class Module1: public sc_module {

public:

  SC_HAS_PROCESS(Module1);
  SubModule2 *submodule2;

  Module1(sc_module_name mn):
    sc_module(mn)
  {
    submodule2 = new SubModule2("submodule2");
    SC_THREAD(thread1);
  }

  int i;
  void thread1()
  {
    int j = 2;

    i++;
    wait(SC_ZERO_TIME);
    j++;
  }
};


class Module2: public sc_module {

public:

  sc_port<Interface> port_;

  SC_HAS_PROCESS(Module2);

  Module2(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(thread1);
  }

  int i;
  void thread1()
  {
    int j = 2;

    i++;
    wait(SC_ZERO_TIME);
    j++;

    port_->write(j);
  }
};

class Module3: public sc_module {

public:

  sc_port<Interface> port_;

  SC_HAS_PROCESS(Module3);

  Module3(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(thread1);
  }

  int i;
  void thread1()
  {
    int j = 2;

    i++;
    wait(SC_ZERO_TIME);
    j++;
    port_->read();
  }
};


class Top: public sc_module {

public:
  Module1 *m1;
  Module2 *m2[2];
  Module3 *m3[2];

  Channel c1;
  Channel c2;

  SC_HAS_PROCESS(Top);

  Top(sc_module_name mn):
    sc_module(mn),
    c1("c1"),
    c2("c2")
  {
    m1 = new Module1("m1");
    m2[0] = new Module2("m21");
    m2[1] = new Module2("m22");
    m3[0] = new Module3("m31");
    m3[1] = new Module3("m32");

    m2[0]->port_.bind(c1);
    m3[0]->port_.bind(c1);

    m2[1]->port_.bind(c2);
    m3[1]->port_.bind(c2);
  }
};

int sc_main(int argc, char **argv)
{
  Top t("top");
  sc_start();

  return 0;
};
