#include "systemc.h"

class Module0: public sc_module {

public:

  sc_port<sc_fifo_out_if<int> > out_port0;

  SC_HAS_PROCESS(Module0);

  Module0(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(thread_m1);
  }

  void thread_m1()
  {
    for(int i = 0; i<20; i++) {
      std::cout << "writing " << i + 10 << std::endl;
      out_port0->write(i+10);
    }
  }
};

class Module1: public sc_module {

public:

  sc_port<sc_fifo_out_if<int> > out_port1;

  Module0 m0;

  SC_HAS_PROCESS(Module1);

  Module1(sc_module_name mn):
    sc_module(mn),
    m0("m0")
  { }

  void before_end_of_elaboration()
  {
    m0.out_port0.bind(out_port1);
  }
};

class Module2: public sc_module {

public:

  sc_port<sc_fifo_out_if<int> > out_port2;

  Module1 m1;

  SC_HAS_PROCESS(Module2);

  Module2(sc_module_name mn):
    sc_module(mn),
    m1("m1")
  {
    SC_THREAD(thread_m2);
  }

  void before_end_of_elaboration()
  {
    m1.out_port1.bind(out_port2);
  }

  void thread_m2()
  {
  }
};

class Module3: public sc_module {

public:

  SC_HAS_PROCESS(Module3);

  sc_port<sc_fifo_in_if<int> > in_port;

  Module3(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(thread_m3);
  }

  void thread_m3()
  {
    while(true) {
      int value = in_port->read();
      std::cout << "reading " << value << std::endl;
    }
  }
};

class ModuleD: public sc_module {

public:
  Module2 left;
  Module2 right;

  sc_port<sc_fifo_out_if<int> > out_port_left_md;
  sc_port<sc_fifo_out_if<int> > out_port_right_md;


  ModuleD(sc_module_name mn):
    sc_module(mn),
    left("left"),
    right("right")
  { }

  void before_end_of_elaboration()
  {
    left.out_port2.bind(out_port_left_md);
    right.out_port2.bind(out_port_right_md);
  }
};

class ModuleE: public sc_module {

public:
  sc_port<sc_fifo_in_if<int> > in_port_left_me;
  sc_port<sc_fifo_in_if<int> > in_port_right_me;

  ModuleE(sc_module_name mn):
    sc_module(mn)
  { }
};

class Up: public sc_module {

public:
  sc_port<sc_fifo_in_if<int> > in_port_up_1;
  sc_port<sc_fifo_in_if<int> > in_port_up_2;

  Up(sc_module_name mn):
    sc_module(mn)
  { }

  void before_end_of_elaboration()
  { }
};

class Down: public sc_module {

public:
  sc_port<sc_fifo_out_if<int> > out_port_down_1;
  sc_port<sc_fifo_out_if<int> > out_port_down_2;

  Down(sc_module_name mn):
    sc_module(mn)
  { }

  void before_end_of_elaboration()
  { }
};

class Testbench: public sc_module {

public:
  SC_HAS_PROCESS(Testbench);

  Module2 m2;
  Module3 m3;

  ModuleD md;
  ModuleE me;

  Up up;
  Down down;

  sc_fifo<int> fifo_;

  sc_fifo<int> fifo_left_;
  sc_fifo<int> fifo_right_;

  sc_fifo<int> fifo_up_down_1;
  sc_fifo<int> fifo_up_down_2;

  Testbench(sc_module_name mn):
    sc_module(mn),
    m2("m2"),
    m3("m3"),
    md("md"),
    me("me"),
    up("up"),
    down("down") {
  }

  void before_end_of_elaboration()
  {
    m2.out_port2.bind(fifo_);
    m3.in_port.bind(fifo_);

    md.out_port_left_md.bind(fifo_left_);
    md.out_port_right_md.bind(fifo_right_);

    me.in_port_left_me.bind(fifo_left_);
    me.in_port_right_me.bind(fifo_right_);

    up.in_port_up_1.bind(fifo_up_down_1);
    up.in_port_up_2.bind(fifo_up_down_2);

    down.out_port_down_1.bind(fifo_up_down_1);
    down.out_port_down_2.bind(fifo_up_down_2);

    up.in_port_up_1.bind(fifo_up_down_1);
    up.in_port_up_2.bind(fifo_up_down_2);
  }
};

int sc_main(int argc, char **argv)
{
  Testbench tb("tb");
  sc_start();

  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
