// bug20180309.cc:
// example "vespa" by A. Dingankar, 3/9/18
// modified to carry some load by RD, 03/09/18
// added to regression test suite by RD, 04/17/18

#include <systemc.h> 

#define LOAD_COUNT	100000
#define ITER_COUNT	100

SC_MODULE(c_Fetch) {
		public:
		sc_fifo_out<int> p_InstrQ;
		int InstrQ;
		int Loop_index;

		void init() {
			}
		void process() {
			InstrQ = (Loop_index % 32) << 27;
			for(int Load=0; Load<LOAD_COUNT; Load++) {Load++;}
			}
		void behavior() {
			init(); 
			wait(10, SC_NS);
			Loop_index = 0;
			while (Loop_index < ITER_COUNT) {
				process(); 
				wait(10, SC_NS);
				p_InstrQ.write(InstrQ);
				wait(1, SC_NS);
				Loop_index++;
				}
		}
		
		SC_CTOR(c_Fetch) {
		        SC_THREAD(behavior);
		}
};

SC_MODULE(c_Decode) {
		public:
		sc_fifo_in<int> p_InstrQ;
		int InstrQ;
		sc_fifo_out<int> p_OpcodeQ;
		int OpcodeQ;
		int Loop_index;

		void init() {
			}
		void process() {
			for(int Load=0; Load<LOAD_COUNT; Load++) {Load++;}
			}
		void behavior() {
			init(); 
			wait(10, SC_NS);
			Loop_index = 0;
			while (1) {
				p_InstrQ.read(InstrQ);
				wait(1, SC_NS);
				process(); 
				wait(10, SC_NS);
				p_OpcodeQ.write(OpcodeQ);
				wait(1, SC_NS);
				Loop_index++;
				}
		}
		
		SC_CTOR(c_Decode) {
		        SC_THREAD(behavior);
		}
};

SC_MODULE(c_Execute) {
		public:
		sc_fifo_in<int> p_OpcodeQ;
		int OpcodeQ;
		sc_fifo_out<int> p_ResultQ;
		int ResultQ;
		int Loop_index;

		void init() {
			}
		void process() {
			for(int Load=0; Load<LOAD_COUNT; Load++) {Load++;}
			}
		void behavior() {
			init(); 
			wait(10, SC_NS);
			Loop_index = 0;
			while (1) {
				p_OpcodeQ.read(OpcodeQ);
				wait(1, SC_NS);
				process(); 
				wait(10, SC_NS);
				p_ResultQ.write(ResultQ);
				wait(1, SC_NS);
				Loop_index++;
				}
		}

		SC_CTOR(c_Execute) {
		        SC_THREAD(behavior);
		}
};

SC_MODULE(c_Writeback) {
		public:
		sc_fifo_in<int> p_ResultQ;
		int ResultQ;
		int Loop_index;

		void init() {
			}
		void process() {
			for(int Load=0; Load<LOAD_COUNT; Load++) {Load++;}
			}
		void behavior() {
			init(); 
			wait(10, SC_NS);
			Loop_index = 0;
			while (1) {
				p_ResultQ.read(ResultQ);
				wait(1, SC_NS);
				process(); 
				wait(10, SC_NS);
				Loop_index++;
				}
		}
		
		SC_CTOR(c_Writeback) {
		        SC_THREAD(behavior);
		}
};

class top : public sc_module {
	public:
		sc_fifo<int> inst_InstrQ;
		sc_fifo<int> inst_OpcodeQ;
		sc_fifo<int> inst_ResultQ;
		c_Fetch inst_Fetch;
		c_Decode inst_Decode;
		c_Execute inst_Execute;
		c_Writeback inst_Writeback;
		
	   top(sc_module_name name) : sc_module(name)
		, inst_InstrQ(1)
		, inst_OpcodeQ(1)
		, inst_ResultQ(1)
		, inst_Fetch("Fetch")
		, inst_Decode("Decode")
		, inst_Execute("Execute")
		, inst_Writeback("Writeback")
		{
		inst_Fetch.p_InstrQ.bind(inst_InstrQ);
		inst_Decode.p_InstrQ.bind(inst_InstrQ);
		inst_Decode.p_OpcodeQ.bind(inst_OpcodeQ);
		inst_Execute.p_OpcodeQ.bind(inst_OpcodeQ);
		inst_Execute.p_ResultQ.bind(inst_ResultQ);
		inst_Writeback.p_ResultQ.bind(inst_ResultQ);
    }
};

int sc_main (int, char *[]) {
	sc_trace_file* Tf; // should create warning, but not crash!

	top top1("Top1");

	sc_start();
//	sc_stop(); // RD
	return 0;
}
