// bug20180723c.cc
// (was port_error.cpp reported by DM 07/23/18)

#include <systemc.h>

class write_if : virtual public sc_interface {
    public:
      virtual void write(char) = 0;
      virtual void reset() = 0;
};
 
class read_if : virtual public sc_interface {
  public:
    virtual void read(char &) = 0;
    virtual int num_available() = 0;
};

class fifo : public sc_channel, public write_if, public read_if {
  public: 
    fifo(sc_module_name name) ;

    void write(char c) ;

    void read(char &c) ;

    void reset();

    int num_available();

  public:
    enum e { max = 2 };
    char data[max];
    int num_elements, first;
    sc_event write_event, read_event;
};

fifo::fifo(sc_module_name name) : 
  sc_channel(name), 
  num_elements(0), 
  first(0) 
{  
}
 
void fifo::write(char c) {
  if (num_elements == max)
    wait(read_event);

  data[(first + num_elements) % max] = c;
  ++ num_elements;
  write_event.notify(SC_ZERO_TIME);
}

void fifo::read(char &c) {
  if (num_elements == 0)
    wait(write_event);

  c = data[first];
  -- num_elements;
  first = (first + 1) % max;
  read_event.notify(SC_ZERO_TIME);
}

void  
fifo::reset() { num_elements = first = 0; }

int  
fifo::num_available() { return num_elements;}
 
SC_MODULE(producer) {
  public:
    sc_port<write_if> out;

    SC_CTOR(producer) {
      SC_THREAD(loop);
    }

    void loop() {
      char c = 'c';

        out->write(c);       
	cout << ">> producer successfully sent: " << c << endl;   
    }
};

SC_MODULE(consumer) {
  public:
    sc_port<read_if> in;

    SC_CTOR(consumer) {
      SC_THREAD(loop);
    }

    void loop() {
      char c;
        in->read(c);
        cout << ">> consumer success received: " << c << endl;
    }
};

SC_MODULE(top) {
public:

  producer prod_inst;
  consumer cons_inst;
  //fifo fifo_inst;	// when instantiated here, port binding analysis succeeds! 

  SC_CTOR(top) :
  fifo_inst("Fifo1"),
  prod_inst("Producer1"),
  cons_inst("Consumer1")
  {
    prod_inst.out.bind(fifo_inst); // port binding in constructor
    cons_inst.in.bind(fifo_inst);
  }
  fifo fifo_inst;	// this late placement appears to be the problem!

  void before_end_of_elaboration(void)
  {
    //prod_inst.out.bind(fifo_inst);
    //cons_inst.in.bind(fifo_inst);
  }
};

int sc_main (int, char *[]) {
 top top1("Top1");
 sc_start();
 return 0;
}
