// run_ahead1.cc: regression test for parallel execution
// (based on parallel_threads1.cc)
// RD, 03/27/19

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define ITERATIONS 10
#define LIGHT_WORKLOAD      50000
#define HEAVY_WORKLOAD   50000000

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

SC_MODULE(Worker1)	// works at 1 ns, 11 ns, 21 ns, ...
{
    unsigned Count;
    unsigned long w;

    void main(void)
    {
	wait(1, SC_NS);
	cout << sc_time_stamp() << ": worker1 starts light work.\n";
	Count++;
	w = Work(LIGHT_WORKLOAD);
	for(int i=0; i<ITERATIONS; i++)
	{
	    wait(10, SC_NS);
	    Count++;
	    w = Work(LIGHT_WORKLOAD);
	}
	cout << sc_time_stamp() << ": worker1 light work done.\n";
	assert(Count == ITERATIONS+1);
	sc_time expected_time(1+10*ITERATIONS, SC_NS);
	assert(sc_time_stamp() == expected_time);
    }

    SC_CTOR(Worker1)
    {
	Count = 0;
	SC_HAS_PROCESS(Worker1);
	SC_THREAD(main);
    }
};

SC_MODULE(Worker2)	// works at 2 ns, 12 ns, 22 ns, ...
{
    unsigned Count;
    unsigned long w;

    void main(void)
    {
	wait(2, SC_NS);
	cout << sc_time_stamp() << ": worker2 starts heavy work.\n";
	Count++;
	w = Work(HEAVY_WORKLOAD);
	for(int i=0; i<ITERATIONS; i++)
	{
	    wait(10, SC_NS);
	    Count++;
	    w = Work(HEAVY_WORKLOAD);
	}
	cout << sc_time_stamp() << ": worker2 heavy work done.\n";
	assert(Count == ITERATIONS+1);
	sc_time expected_time(2+10*ITERATIONS, SC_NS);
	assert(sc_time_stamp() == expected_time);
    }

    SC_CTOR(Worker2)
    {
	Count = 0;
	SC_HAS_PROCESS(Worker2);
	SC_THREAD(main);
    }
};

SC_MODULE(TOP)
{
    Worker1 w1;
    Worker2 w2;

    SC_CTOR(TOP)
    : w1("w1"), w2("w2")
    {
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    sc_time expected_time(2+10*ITERATIONS, SC_NS);

    cout << sc_time_stamp() << ": run_ahead1 starts\n";
    sc_start();
    cout << sc_time_stamp() << ": run_ahead1 done.\n";
    assert(top.w1.Count == ITERATIONS+1);
    assert(top.w2.Count == ITERATIONS+1);
    assert(sc_time_stamp() == expected_time);
    return 0;
}

// EOF
