// varref.cpp: crash in resolving local reference variables (01/13/19, RD).
// - extracted from particle_simulation.cpp
// - reproduce crash: risc -vv -Wall -o varref_ooo varref.cpp

#include <stdio.h>
#include <assert.h>
#include "systemc.h"

#define NUM_DUT 1
#define NUM_ROWS 1
#define NUM_COLS 1
#define NUM_PARTICLES 100
#define MAX_PARTICLES 100
#define density 0.0005
#define mass    0.01
#define cutoff  0.01
#define min_r   (cutoff/100)
#define dt      0.0005

typedef struct Particle_t
{
  double 	x, y;
  double 	vx, vy;
  double	ax, ay;
  unsigned	color;
  int		id;

  inline friend std::ostream& operator<<(
	std::ostream& os, const Particle_t& co)
  { os << co.x << ' ' << co.y << '\n';
    return os;
  }
} particle_t;

unsigned const RED   = 0xff0000;
unsigned const BLUE  = 0x0000ff;
unsigned const WHITE = 0xffffff;
const int INVALID = -1;
const particle_t pINVALID = {0,0,0,0,0,0,WHITE,INVALID};
enum {NORTH, EAST, SOUTH, WEST};

class i_particle_transceiver: virtual public sc_interface
{
    public:
	virtual std::vector<particle_t> read() = 0;
	virtual void write(const std::vector<particle_t>) = 0;
};

class c_particle_double_handshake_channel:
	public sc_channel, public i_particle_transceiver
{
    public:
	c_particle_double_handshake_channel(sc_module_name name):
		sc_channel(name)
	{
	    v = false;
	    w = false;
	}

	~c_particle_double_handshake_channel()
	{ }

	std::vector<particle_t> read()
	{
	    if (!v)
	    {
		w = true;
		wait(req);
		w = false;
	    }
	    std::vector<particle_t> temp = tmpd;
	    v = false;
	    ack.notify(SC_ZERO_TIME);
	    wait(ack);
	    return temp;
	}

	void write(const std::vector<particle_t> X)
	{
	    tmpd = X;
	    v = true;
	    if (w)
	    {
		req.notify(SC_ZERO_TIME);
	    }
	    wait(ack);
	}
    private:
	sc_event req;
	sc_event ack;
	bool v, w;
	std::vector<particle_t> tmpd;
};

SC_MODULE(DUT)
{
    sc_port<i_particle_transceiver> CoordsControl;
    sc_port<i_particle_transceiver> CoordsNorth;
    sc_port<i_particle_transceiver> CoordsEast;
    sc_port<i_particle_transceiver> CoordsSouth;
    sc_port<i_particle_transceiver> CoordsWest;

    void main(void);
    void apply_force(int p, int n);
    void advanced_move(int pIndex);

    DUT(sc_module_name name_, int Row_ = 1, int Col_ = 1):
	sc_module(name_),
	Row(Row_), Col(Col_),
	TileSize(sqrt(density * NUM_PARTICLES) / sqrt(NUM_DUT)),
	LeftBound((Col-1)*TileSize),
	RightBound(LeftBound + TileSize),
	BottomBound((NUM_ROWS-Row)*TileSize),
	TopBound(BottomBound + TileSize)
    {
	SC_HAS_PROCESS(DUT);
	SC_THREAD(main);
    }

    private:
	const int Row, Col;
	const double TileSize;
	const double LeftBound;
	const double RightBound;
	const double BottomBound;
	const double TopBound;
	particle_t particles[MAX_PARTICLES];
	int nParticlesInUse;
	std::vector<std::vector<particle_t> > fugitives;
};

void DUT::main(void)
{
	particle_t p;

	std::vector<particle_t> pN, pE, pS, pW;
	fugitives.push_back(pN);
	assert(0 == NORTH);
	fugitives.push_back(pE);
	assert(1 == EAST);
	fugitives.push_back(pS);
	assert(2 == SOUTH);
	fugitives.push_back(pW);
	assert(3 == WEST);
	nParticlesInUse = 0;
	std::vector<particle_t> pp;
	pp = CoordsControl->read();
	for(std::vector<particle_t>::iterator iter = pp.begin(); iter != pp.end(); iter++) {
	    p = *iter;
	    if (p.id == INVALID)
		break;
	    particles[nParticlesInUse++] = p;
	    assert(nParticlesInUse <= MAX_PARTICLES);
	    assert(LeftBound <= p.x);
	    assert(RightBound >= p.x);
	    assert(BottomBound <= p.y);
	    assert(TopBound >= p.y);
	}
	for (int i=nParticlesInUse; i<MAX_PARTICLES; i++)
	{
	    particles[i] = pINVALID;
	}
	for(int step = 0; step < 10; step++)
	{
	    for(int i=0; i<nParticlesInUse; i++)
	    {
		if (particles[i].id == INVALID)
		    continue;
		for (int j=0; j<nParticlesInUse; j++)
		{
		    if (particles[j].id != INVALID)
		    {
			apply_force(i, j);
		    }
		}
	    }
	    for(int i = 0; i <nParticlesInUse; i++)
	    {
		if (particles[i].id != INVALID)
		{
		  advanced_move(i);
		}
	    }
	}

	int count = 0;
	for(int i = 0; i< MAX_PARTICLES; i++){
	    if (particles[i].id != INVALID)
		count++;
	    assert(i < nParticlesInUse || particles[i].id == INVALID);
	}
	assert(nParticlesInUse >= count);
	assert(nParticlesInUse <= MAX_PARTICLES);
}

void DUT::apply_force(int p, int n)
{
    particle_t &particle = particles[p];	// crash was here!
    particle_t &neighbor = particles[n];	// problem is here!
//  #define particle particles[p]		// work-around is this
//  #define neighbor particles[n]		// work-around is this

    assert(particle.id != INVALID);
    assert(neighbor.id != INVALID);
    double dx = neighbor.x - particle.x;
    double dy = neighbor.y - particle.y;
    double r2 = dx * dx + dy * dy;
    if( r2 > cutoff*cutoff )
	return;
    r2 = fmax( r2, min_r*min_r );
    double r = sqrt( r2 );
    double coef = ( 1 - cutoff / r ) / r2 / mass;
    particle.ax += coef * dx;
    particle.ay += coef * dy;

    if (particle.ax*particle.ax + particle.ay*particle.ay > 10000000)
    {
	particle.ax = 0;
	particle.ay = 0;
    }
}

void DUT::advanced_move(int pIndex)
{
    particle_t &p = particles[pIndex];	// crash was here!
//  #define p particles[pIndex]		// work-around is this
    assert(p.id != INVALID);
    p.vx += p.ax * dt;
    p.vy += p.ay * dt;
    p.x  += p.vx * dt;
    p.y  += p.vy * dt;
    if (p.x < LeftBound)
    {
	if (Col != 1)
	{
	    p.color = RED;
	    fugitives[WEST].push_back(p);
	    p = pINVALID;
	    return;
	}
	else
	{
	    p.x  = -p.x;
	    p.vx = -p.vx;
	    p.color = BLUE;
	}
    }
    if (p.x > RightBound)
    {
	if (Col != NUM_COLS)
	{
	    p.color = RED;
	    fugitives[EAST].push_back(p);
	    p = pINVALID;
	    return;
	}
	else
	{
	    p.x = 2*RightBound-p.x;
	    p.vx= -p.vx;
	    p.color = BLUE;
	}
    }
    if (p.y < BottomBound)
    {
	if (Row != NUM_ROWS)
	{
	    p.color = RED;
	    fugitives[SOUTH].push_back(p);
	    p = pINVALID;
	    return;
	}
	else
	{
	    p.y  = -p.y;
	    p.vy = -p.vy;
	    p.color = BLUE;
	}
    }
    if (p.y > TopBound)
    {
	if (Row != 1)
	{
	    p.color = RED;
	    fugitives[NORTH].push_back(p);
	    p = pINVALID;
	    return;
	}
	else
	{
	    p.y = 2*TopBound-p.y;
	    p.vy = -p.vy;
	    p.color = BLUE;
	}
    }
}

SC_MODULE(TOP)
{
    c_particle_double_handshake_channel   c_north;
    c_particle_double_handshake_channel   c_east;
    c_particle_double_handshake_channel   c_south;
    c_particle_double_handshake_channel   c_west;
    c_particle_double_handshake_channel   c_1;
    DUT		dut;

    SC_CTOR(TOP)
    :c_north("c_north"), c_east("c_east"), c_south("c_south"), c_west("c_west"),
	c_1("c_1"), dut("dut")
    {
	dut(c_1, c_north, c_east, c_south, c_west);
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("varref starts.\n");
    sc_start();
    printf("varref ends.\n");
    return 0;
}

// EOF
