#ifndef GENERIC_PAYLOAD_ATTRIBUTES_H_INCLUDED_
#define GENERIC_PAYLOAD_ATTRIBUTES_H_INCLUDED_

#include "rose.h"
#include "port_call_path.h"

#include<set>
#include<unordered_map>
#include<list>

namespace risc {

namespace sg {

/* 
  the reason why we need these three attributes:
  a generic payload (GP) object is used both in an initiator and
  refered in the customized_b_transport. So, we somehow need a 
  way to know that how the initiator field data pointed to by the GP 
  object is used (read, write) in the customized_b_transport
*/

class OriginalDataAttribute: public AstAttribute {

public:

  // for the new rose release
  virtual AstAttribute::OwnershipPolicy getOwnershipPolicy() const ROSE_OVERRIDE
  {
    return CONTAINER_OWNERSHIP;
  }

  /**
   * \brief SgVariableSymbol for the original variable
   */
  SgVariableSymbol* data_symbol_;
};

class MappedDataAttribute: public AstAttribute {
public:

  // for the new rose release
  virtual AstAttribute::OwnershipPolicy getOwnershipPolicy() const ROSE_OVERRIDE
  {
    return CONTAINER_OWNERSHIP;
  }

  /**
   * \brief SgVariableSymbol for the alias variable
   */
  SgVariableSymbol* data_symbol_;
};

//for now, can be dmi or generic payload
class Tlm2ContainerOfMappedDataAttribute: public AstAttribute {
public:

  // for the new rose release
  virtual AstAttribute::OwnershipPolicy getOwnershipPolicy() const ROSE_OVERRIDE
  {
    return CONTAINER_OWNERSHIP;
  }

  /**
   * \brief SgVariableSymbol for the generic payload
   */
  SgVariableSymbol* container_;
};

class MappedContainersOfContainerAttribute: public AstAttribute {
public:

  // for the new rose release
  virtual AstAttribute::OwnershipPolicy getOwnershipPolicy() const ROSE_OVERRIDE
  {
    return CONTAINER_OWNERSHIP;
  }

  /**
   * \brief If a container (say, generic payload) is approached with socket S, then,
   *        its mapped containers are those in the list
   */
  std::unordered_map<SgVariableSymbol*, std::list<SgVariableDefinition*> > symbols_sockets_;
};

class DmiSocketAttribute: public AstAttribute {
public:

  // for the new rose release
  virtual AstAttribute::OwnershipPolicy getOwnershipPolicy() const ROSE_OVERRIDE
  {
    return CONTAINER_OWNERSHIP;
  }

  /**
   * \brief The socket that calls the b_transport
   */
  SgVariableDefinition* socket_;
};

//currently, the methods include b_transport and get_direct_mem_ptr
class SocketMethodParameterReferenceAttribute: public AstAttribute {

public:

  // for the new rose release
  virtual AstAttribute::OwnershipPolicy getOwnershipPolicy() const ROSE_OVERRIDE
  {
    return CONTAINER_OWNERSHIP;
  }

  /**
   * \brief The container is approached by what socket
   */
  std::unordered_map<SgVariableSymbol*, SgVariableDefinition* > reference_and_socket_;
};

} // end of namespace sg

} // end of namespace risc

#endif /* GENERIC_PAYLOAD_ATTRIBUTES_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
